from __future__ import annotations

from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.testing.common import create_sheet
from xlviews.testing.sheet_frame.base import MultiColumn

if __name__ == "__main__":
    sheet = create_sheet()
    fc = MultiColumn(sheet, column=2, style=True)
    sf = fc.sf

    ax = Axes(row=13, column=2, chart_type=ChartType.xlXYScatter)
    df = sf.groupby("s")  # .agg(include_sheetname=True)
    # for key, s in df.iterrows():
    #     print(key, s["x"], s["y"])
    #     ax.add_series(s["x"], s["y"], label=f"{key}")
    # x = sf.range("x")
    # y = sf.range("y")
    # label = sf.first_range("a")
    # ax.add_series(x, y, label=label)
    # ax.add_series(x.get_address(include_sheetname=True), y)
    # ax.chart.api[1].ChartTitle.Text = sheet_module["A1"].api
    # ax.chart.api[1].ChartTitle.Text = "=a1"

    # ax.xlabel = sf.range("x", -1)
    # ax.ylabel = sf.range("y", -1)

    # gr = sf.groupby(None)

    # gr = sf.groupby(["a", "b", "c"])
    # key = ("c", "t")
    # x = gr.range("x", key)
    # y = gr.range("y", key)
    # label = gr.first_range("b", key)
    # ax.add_series(x, y, label=label)

    # for x, y in zip(gr.ranges("x"), gr.ranges("y"), strict=True):
    #     ax.add_series(x, y)

    # list(gr.keys())

    # list(gr.first_ranges("a"))

    # ax.tight_layout()
    # ax.set_style()
    # ax.set_legend(loc=(0, 0))

    # g = data.groupby("a")
    # key = ("u",)
    # x = data.range("x", g[key])
    # y = data.range("y", g[key])
    # x
    # s = ax.add_series(
    #     x,
    #     y,
    #     # label=data.range("a", g["v"])[0],
    #     chart_type=ChartType.xlXYScatterLines,
    # )
    # ax.xlabel = data.range("x", 0)
    # ax.ylabel = data.range("y", 0)
    # ax.title = "=" + data.range("a", 0).get_address(include_sheetname=True)



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.dataframes.heat_frame import Colorbar, facet
from xlviews.testing.common import create_sheet
from xlviews.testing.heat_frame.common import HeatFrameContainer
from xlviews.testing.sheet_frame.pivot import Pivot

if TYPE_CHECKING:
    from pandas import DataFrame

    from xlviews.dataframes.sheet_frame import SheetFrame


class FacetParent(Pivot):
    pass


class Facet(HeatFrameContainer):
    @classmethod
    def dataframe(cls, sf: SheetFrame) -> DataFrame:
        return sf.pivot_table("u", ["B", "Y"], ["A", "X"], "mean", formula=True)

    def init(self) -> None:
        super().init()


if __name__ == "__main__":
    sheet = create_sheet()

    fc = FacetParent(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)
    fc = Facet(sf)
    fc.sf.set_adjacent_column_width(1)
    fc.sf.range.impl.number_format = "0"
    fc.sf.autofit()

    df = sf.pivot_table("u", ["B", "Y"], ["A", "X"], "mean", formula=True)
    g = facet(2, 18, df, "B", "A")
    g[0, 0].autofit()
    g[0, 0].set_adjacent_column_width(1)
    g[0, 1].autofit()
    g[0, 1].set_adjacent_column_width(1)
    print(g)

    print(sf.agg(None, "u"))
    rng = sf.get_range("u")
    cell = g[0, -1].get_adjacent_cell()

    cb = Colorbar(cell.row + 1, cell.column, 6)
    cb.set(vmin=rng, vmax=rng).autofit()
    for x in g.flat:
        cb.apply(x.range)

    df = sf.pivot_table("u", ["B", "Y", "y"], ["A", "X", "x"], formula=True)
    g = facet(20, 18, df, "B", "A")
    g[0, 0].autofit()
    g[0, 0].set_adjacent_column_width(1)
    g[0, 1].autofit()
    g[0, 1].set_adjacent_column_width(1)
    print(g)

    print(sf.agg(None, "u"))
    rng = sf.get_range("u")
    cell = g[0, -1].get_adjacent_cell()

    cb = Colorbar(cell.row + 1, cell.column, 6)
    cb.set(vmin=rng, vmax=rng).autofit()
    for x in g.flat:
        cb.apply(x.range)



import pytest
from xlwings import Sheet

from xlviews.dataframes.groupby import groupby
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import WideColumn

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return WideColumn(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 3
    assert sf.column == 29
    assert sf.sheet.name == sheet_module.name
    assert sf.index.nlevels == 2
    assert sf.columns.nlevels == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 5


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["x", "y"]


def test_contains(sf: SheetFrame):
    assert "x" not in sf
    assert "a" in sf
    assert "u" in sf


def test_iter(sf: SheetFrame):
    assert list(sf) == ["a", "b", "u", "v"]


# @pytest.mark.parametrize(
#     ("column", "index"),
#     [
#         ("a", 31),
#         ("b", 32),
#         (["y", "b"], [30, 32]),
#         ("u", (33, 35)),
#         ("v", (36, 39)),
#         (("v", 0), 36),
#         (("v", 3), 39),
#         (["x", "a", "u", ("v", 0)], [29, 31, (33, 35), 36]),
#     ],
# )
# def test_index(sf: SheetFrame, column, index):
#     assert sf.index_past(column) == index


# @pytest.mark.parametrize("column", ["z", ("u", -1)])
# def test_index_error(sf: SheetFrame, column):
#     with pytest.raises(ValueError, match=".* is not in list"):
#         sf.index_past(column)


# @pytest.mark.parametrize(
#     ("column", "offset", "address"),
#     [
#         ("x", -1, "$AC$3"),
#         ("y", 0, "$AD$4"),
#         ("a", None, "$AE$4:$AE$8"),
#         (("u", 0), -1, "$AG$2:$AG$3"),
#         (("u", 2), 0, "$AI$4"),
#         (("v", 0), None, "$AJ$4:$AJ$8"),
#         ("u", -1, "$AG$2:$AI$3"),
#         ("u", 0, "$AG$4:$AI$4"),
#         ("v", None, "$AJ$4:$AM$8"),
#     ],
# )
# def test_range_column(sf: SheetFrame, column, offset, address):
#     assert sf.range(column, offset).get_address() == address


@pytest.mark.parametrize(
    ("by", "v1", "v2"),
    [
        ("x", [(4, 5), (8, 8)], [(6, 7)]),
        ("y", [(4, 4), (6, 6), (8, 8)], [(5, 5), (7, 7)]),
    ],
)
def test_groupby(sf: SheetFrame, by, v1, v2):
    g = groupby(sf, by)
    assert len(g) == 2
    keys = list(g.keys())
    assert g[keys[0]] == v1
    assert g[keys[1]] == v2


def test_groupby_list(sf: SheetFrame):
    g = groupby(sf, ["x", "y"])
    assert len(g) == 4
    assert g[("i", "k")] == [(4, 4), (8, 8)]
    assert g[("i", "l")] == [(5, 5)]
    assert g[("j", "k")] == [(6, 6)]
    assert g[("j", "l")] == [(7, 7)]



from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings import Range as RangeImpl
from xlwings.constants import LineStyle

from xlviews.colors import rgb
from xlviews.core.address import reference
from xlviews.core.range import Range
from xlviews.core.range_collection import RangeCollection

from .style import get_line_style, get_marker_style

if TYPE_CHECKING:
    from typing import Any

    from xlwings import Sheet

    from .axes import Axes


class Series:
    axes: Axes
    api: Any
    label: str

    def __init__(
        self,
        axes: Axes,
        x: Any,
        y: Any | None = None,
        label: str | tuple[int, int] | Range | RangeImpl = "",
        chart_type: int | None = None,
        sheet: Sheet | None = None,
    ) -> None:
        self.axes = axes
        self.api = axes.chart.api[1].SeriesCollection().NewSeries()
        self.label = label if isinstance(label, str) else reference(label, sheet)
        self.name = self.label

        if chart_type is not None:
            self.chart_type = chart_type

        if y is not None:
            self.x = x
            self.y = y

        else:
            self.y = x

    @property
    def name(self) -> str:
        return str(self.api.Name)

    @name.setter
    def name(self, name: str) -> None:
        self.api.Name = name

    @property
    def chart_type(self) -> int:
        return self.api.ChartType  # type: ignore

    @chart_type.setter
    def chart_type(self, chart_type: int) -> None:
        self.api.ChartType = chart_type

    @property
    def x(self) -> tuple:
        return self.api.XValues  # type: ignore

    @x.setter
    def x(self, x: Any) -> None:
        if isinstance(x, Range | RangeImpl | RangeCollection):
            self.api.XValues = x.api
        else:
            self.api.XValues = x

    @property
    def y(self) -> tuple:
        return self.api.Values  # type: ignore

    @y.setter
    def y(self, y: Any) -> None:
        if isinstance(y, Range | RangeImpl | RangeCollection):
            self.api.Values = y.api
        else:
            self.api.Values = y

    def delete(self) -> None:
        self.api.Delete()

    def set(
        self,
        marker: str | None = "o",
        line: str | None = "-",
        color: str | int | tuple[int, int, int] = "black",
        size: int = 5,
        weight: float = 2,
        alpha: float = 0,
        **kwargs,
    ) -> None:
        if line is None:
            weight = min(size / 4, weight)
            line_alpha = alpha / 2
        else:
            line_alpha = alpha

        set_marker(self.api, get_marker_style(marker), size)
        set_fill(self.api, rgb(color), alpha)
        set_line(self.api, rgb(color), get_line_style(line), weight, line_alpha)


def set_marker(api: Any, style: int, size: int) -> None:
    api.MarkerStyle = style
    api.MarkerSize = size


def set_fill(api: Any, color: int, alpha: float) -> None:
    api.Format.Fill.Visible = True
    api.Format.Fill.BackColor.RGB = color
    api.Format.Fill.Transparency = alpha
    api.Format.Fill.ForeColor.RGB = color


def set_line(api: Any, color: int, style: int, weight: float, alpha: float) -> None:
    api.Border.LineStyle = LineStyle.xlContinuous
    api.Format.Line.Visible = True
    api.Format.Line.Weight = weight
    api.Format.Line.Transparency = alpha
    api.Format.Line.ForeColor.RGB = color
    api.Border.LineStyle = style



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.testing import is_app_available


@pytest.mark.parametrize(
    ("name", "value"),
    [
        ("ChartType.xlXYScatter", -4169),
        ("BordersIndex.EdgeTop", 8),
        ("Bottom", -4107),
        ("Center", -4108),
        ("Left", -4131),
        ("None", -4142),
        ("Right", -4152),
        ("Top", -4160),
    ],
)
def test_constant(name: str, value: int):
    from xlviews.utils import constant

    assert constant(name) == value
    assert constant(*name.split(".")) == value


@pytest.mark.parametrize(
    ("name", "color"),
    [
        ("black", 0),
        ("red", 2**8 - 1),
        ("green", 32768),
        ("blue", 256 * 256 * 255),
        ("white", 2**24 - 1),
        ("aqua", 16776960),
        ("gray", 8421504),
        ("indigo", 8519755),
        ("lime", 65280),
        ("pink", 13353215),
        ((10, 10, 10), 10 + 10 * 256 + 10 * 256 * 256),
        (100, 100),
    ],
)
def test_rgb(name, color):
    from xlviews.utils import rgb

    assert rgb(name) == color
    if isinstance(name, tuple):
        assert rgb(*name) == color


@pytest.mark.parametrize("name", ["invalid", (1, "x", "y")])
def test_rgb_error(name):
    from xlviews.utils import rgb

    with pytest.raises(ValueError, match="Invalid color format"):
        rgb(name)


@pytest.mark.parametrize(
    ("columns", "lst"),
    [
        ("B", ["B"]),
        (["A", "C"], ["A", "C"]),
        (":B", ["A", "B"]),
        (["::B", "C"], ["A", "C"]),
    ],
)
def test_iter_columns(columns, lst):
    from xlviews.utils import iter_columns

    df = DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], columns=["A", "B", "C"])
    assert list(iter_columns(df, columns)) == lst


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_validate_list(sheet: Sheet):
    from xlviews.utils import add_validate_list

    rng = sheet.range("a1")
    add_validate_list(rng, [1, 2, 3], 2)
    assert rng.value == 2

    assert rng.api.Validation.Type == 3
    assert rng.api.Validation.Operator == 3
    assert rng.api.Validation.Formula1 == "1,2,3"



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.facet import FacetParent

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc_parent(sheet_module: Sheet):
    return FacetParent(sheet_module)


@pytest.fixture(scope="module")
def df(fc_parent: FacetParent):
    return fc_parent.df



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.chart.axes import Axes
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    ("pos", "left", "top"),
    [("right", 312, 18), ("inside", 134, 66), ("bottom", 52, 90)],
)
def test_set_first_position(sheet: Sheet, pos: str, left: float, top: float):
    from xlviews.chart.axes import (
        FIRST_POSITION,
        clear_first_position,
        set_first_position,
    )

    df = DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])
    sf = SheetFrame(2, 2, data=df, sheet=sheet)

    set_first_position(sf, pos)
    assert FIRST_POSITION["left"] == left
    assert FIRST_POSITION["top"] == top
    clear_first_position()


@pytest.mark.parametrize(
    ("args", "expected"),
    [((10, 20), (10, 20)), ((None, None), (50, 50)), ((0, None), (50, 50))],
)
def test_chart_position(sheet: Sheet, args, expected):
    from xlviews.chart.axes import chart_position

    assert chart_position(sheet, *args) == expected


def test_chart_position_from_cell(sheet: Sheet):
    axes = Axes(sheet=sheet, row=5, column=10)
    assert axes.chart.left == 9 * sheet.cells(1, 1).width
    assert axes.chart.top == 4 * sheet.cells(1, 1).height


def test_chart_position_from_chart(sheet: Sheet):
    from xlviews.chart.axes import clear_first_position

    clear_first_position()

    a = Axes(sheet=sheet)
    assert a.chart.left == 50
    assert abs(a.chart.top - 50) <= 0.5

    b = Axes(sheet=sheet)
    assert b.chart.left == a.chart.left + a.chart.width

    c = Axes(0, None, width=200, sheet=sheet)
    assert c.chart.left == a.chart.left
    assert abs(c.chart.top - (a.chart.top + a.chart.height)) <= 0.5

    d = Axes(sheet=sheet)
    assert abs(d.chart.left - (c.chart.left + 200)) <= 0.5
    assert abs(d.chart.top - c.chart.top) <= 0.5



# from __future__ import annotations

# from typing import TYPE_CHECKING

# from xlwings.constants import ChartType

# from xlviews.dataframes.sheet_frame import SheetFrame

# from .axes import Axes

# if TYPE_CHECKING:
#     from xlwings import Range, Sheet

#     from xlviews.chart.series import Series
#     from xlviews.core.range_collection import RangeCollection
#     from xlviews.dataframes.groupby import GroupBy


# def get_range(
#     data: SheetFrame | GroupBy,
#     column: str,
#     key: str | tuple | None = None,
# ) -> Range | RangeCollection:
#     if isinstance(data, SheetFrame):
#         return data.range(column)

#     if isinstance(key, str):
#         key = (key,)

#     return data.range(column, key or ())


# def get_label(
#     data: SheetFrame | GroupBy,
#     column: str,
#     key: str | tuple | None = None,
# ) -> Range:
#     if isinstance(data, SheetFrame):
#         return data.first_range(column)

#     if isinstance(key, str):
#         key = (key,)

#     return data.first_range(column, key or ())


# def plot(
#     data: SheetFrame | GroupBy,
#     x: str,
#     y: str | None = None,
#     *,
#     key: str | tuple | None = None,
#     ax: Axes | None = None,
#     label: str | tuple[int, int] | Range = "",
#     chart_type: int | None = None,
#     sheet: Sheet | None = None,
# ) -> Series:
#     ct = ChartType.xlXYScatter if chart_type is None else chart_type
#     ax = ax or Axes(chart_type=ct)

#     xrng = get_range(data, x, key)
#     yrng = get_range(data, y, key) if y else None

#     if isinstance(label, str):
#         label = get_label(data, label, key)

#     return ax.add_series(xrng, yrng, label=label, chart_type=chart_type, sheet=sheet)



def test_marker_palette():
    from xlviews.chart.style import marker_palette

    mp = marker_palette(10)
    assert mp[0] == "o"
    assert mp[-2] == "*"
    assert mp[-1] == "o"


def test_marker_style_int():
    from xlviews.chart.style import get_marker_style

    assert get_marker_style(1) == 1


def test_line_style_int():
    from xlviews.chart.style import get_line_style

    assert get_line_style(1) == 1



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xlwings import Range as RangeImpl
    from xlwings import Sheet

    from xlviews.core.range import Range


@cache
def index_to_column_name(n: int) -> str:
    """Return the Excel column name from an integer.

    Examples:
        >>> index_to_column_name(1)
        'A'
        >>> index_to_column_name(26)
        'Z'
        >>> index_to_column_name(27)
        'AA'
        >>> index_to_column_name(731)
        'ABC'
    """
    name = ""
    while n > 0:
        n -= 1
        name = chr(n % 26 + 65) + name
        n //= 26

    return name


@cache
def column_name_to_index(col: str) -> int:
    """Return the index from an Excel column name.

    Examples:
        >>> column_name_to_index("A")
        1
        >>> column_name_to_index("Z")
        26
        >>> column_name_to_index("AA")
        27
        >>> column_name_to_index("ABC")
        731
    """
    index = 0
    for char in col:
        index = index * 26 + (ord(char) - ord("A") + 1)

    return index


def reference(
    cell: str | tuple[int, int] | Range | RangeImpl,
    sheet: Sheet | None = None,
) -> str:
    """Return a reference to a cell with sheet name for chart."""
    if isinstance(cell, str):
        return cell

    if isinstance(cell, tuple):
        if sheet is None:
            raise ValueError("`sheet` is required when `cell` is a tuple")

        cell = sheet.range(*cell)

    return "=" + cell.get_address(include_sheetname=True)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xlwings
from pandas import DataFrame, Index, MultiIndex

from xlviews.colors import rgb
from xlviews.config import rcParams
from xlviews.core.formula import aggregate
from xlviews.core.range import Range
from xlviews.core.style import (
    set_alignment,
    set_border,
    set_color_scale,
    set_font,
    set_number_format,
)
from xlviews.utils import suspend_screen_updates

from .sheet_frame import SheetFrame
from .style import set_heat_frame_style

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterator
    from typing import Any, Literal, Self

    from numpy.typing import NDArray
    from pandas import Index
    from xlwings import Sheet


class HeatFrame(SheetFrame):
    index: Index
    columns: Index
    range: Range

    @suspend_screen_updates
    def __init__(
        self,
        row: int,
        column: int,
        data: DataFrame,
        sheet: Sheet | None = None,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
    ) -> None:
        data = clean_data(data)

        super().__init__(row, column, data, sheet)

        self.columns = data.columns

        start = self.row + 1, self.column + 1
        end = start[0] + self.shape[0] - 1, start[1] + self.shape[1] - 1
        self.range = Range(start, end, self.sheet)

        set_heat_frame_style(self)
        self.set(vmin, vmax)

    def set(
        self,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
    ) -> Self:
        rng = self.range

        if vmin is None:
            vmin = aggregate("min", rng)
        if vmax is None:
            vmax = aggregate("max", rng)

        set_color_scale(rng, vmin, vmax)
        return self

    def colorbar(
        self,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
        label: str | None = None,
        autofit: bool = False,
    ) -> Colorbar:
        row = self.row + 1
        column = self.column + self.shape[1] + 2
        length = self.shape[0]

        if vmin is None:
            vmin = self.range
        if vmax is None:
            vmax = self.range

        cb = Colorbar(row, column, length, sheet=self.sheet)
        cb.set(vmin, vmax, label, autofit)
        return cb


def clean_data(data: DataFrame) -> DataFrame:
    data = data.copy()

    if isinstance(data.columns, MultiIndex):
        data.columns = data.columns.droplevel(list(range(1, data.columns.nlevels)))

    if isinstance(data.index, MultiIndex):
        data.index = data.index.droplevel(list(range(1, data.index.nlevels)))

    data.index.name = None

    return data


class Colorbar:
    start: int
    end: int
    offset: int
    orientation: Literal["vertical", "horizontal"] = "vertical"
    sheet: Sheet
    range: Range

    def __init__(
        self,
        row: int,
        column: int,
        length: int,
        orientation: Literal["vertical", "horizontal"] = "vertical",
        sheet: Sheet | None = None,
    ) -> None:
        self.sheet = sheet or xlwings.sheets.active
        self.orientation = orientation

        if orientation == "vertical":
            self.start = row
            self.end = row + length - 1
            self.offset = column
            self.range = Range((self.start, column), (self.end, column), self.sheet)

        else:
            self.start = column
            self.end = column + length - 1
            self.offset = row
            self.range = Range((row, self.start), (row, self.end), self.sheet)

    def set(
        self,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
        label: str | None = None,
        autofit: bool = False,
    ) -> Self:
        if vmin is not None:
            self.vmin = vmin
        if vmax is not None:
            self.vmax = vmax
        if label is not None:
            self.label = label

        self.draw()

        if autofit:
            self.autofit()

        return self

    @property
    def vmin(self) -> Range:
        i = -1 if self.orientation == "vertical" else 0
        return self.range[i]

    @property
    def vmax(self) -> Range:
        i = 0 if self.orientation == "vertical" else -1
        return self.range[i]

    @vmin.setter
    def vmin(self, value: float | str | Range | list[Range]) -> None:
        if isinstance(value, Range | list):
            func = "min" if len(value) > 1 else None
            value = aggregate(func, value, formula=True)

        self.vmin.value = value

    @vmax.setter
    def vmax(self, value: float | str | Range | list[Range]) -> None:
        if isinstance(value, Range | list):
            func = "max" if len(value) > 1 else None
            value = aggregate(func, value, formula=True)

        self.vmax.value = value

    @property
    def label(self) -> Range:
        offset = (-1, 0) if self.orientation == "vertical" else (0, 1)
        return self.vmax.offset(*offset)

    @label.setter
    def label(self, label: str | None) -> None:
        rng = self.label
        rng.value = label
        set_font(rng, bold=True, size=rcParams["frame.font.size"])
        set_alignment(rng, horizontal_alignment="center")

    def draw(self) -> None:
        rng = self.range
        set_color_scale(rng, self.vmin, self.vmax)
        set_font(rng, color=rgb("white"), size=rcParams["frame.font.size"])
        set_alignment(rng, horizontal_alignment="center")
        ec = rcParams["heat.border.color"]
        set_border(rng, edge_weight=2, edge_color=ec, inside_weight=0)

        vmin = self.vmin.get_address()
        vmax = self.vmax.get_address()

        n = self.end - self.start - 1
        for i in range(n):
            value = f"={vmax}+{i + 1}*({vmin}-{vmax})/{n + 1}"
            if self.orientation == "vertical":
                rng = self.sheet.range(self.start + i + 1, self.offset)
            else:
                rng = self.sheet.range(self.offset, self.start + i + 1)

            rng.value = value
            set_font(rng, size=4)
            set_number_format(rng, "0")

    def apply(self, rng: Range) -> None:
        set_color_scale(rng, self.vmin, self.vmax)

    def autofit(self) -> Self:
        if self.orientation == "vertical":
            start = (self.start - 1, self.offset)
            end = (self.end, self.offset)
        else:
            start = (self.offset, self.start)
            end = (self.offset, self.end + 1)

        self.sheet.range(start, end).autofit()
        return self

    def set_adjacent_column_width(self, width: float, offset: int = 1) -> None:
        """Set the width of the adjacent empty column."""
        if self.orientation == "vertical":
            self.range.offset(0, 1).impl.column_width = width
        else:
            self.range.last_cell.offset(0, 2).impl.column_width = width


def facet(
    row: int,
    column: int,
    data: DataFrame,
    index: str | list[str] | None = None,
    columns: str | list[str] | None = None,
    padding: tuple[int, int] = (1, 1),
) -> NDArray:
    frames = []
    for row_, isub in iterrows(data.index, index, row, padding[0] + 1):
        frames.append([])
        for column_, csub in iterrows(data.columns, columns, column, padding[1] + 1):
            sub = xs(data, isub, csub)
            frame = HeatFrame(row_, column_, sub)
            frames[-1].append(frame)

    return np.array(frames)


def iterrows(
    index: Index,
    levels: str | list[str] | None,
    offset: int = 0,
    padding: int = 0,
) -> Iterator[tuple[int, dict[Hashable, Any]]]:
    if levels is None:
        yield offset, {}
        return

    if isinstance(levels, str):
        levels = [levels]

    if levels:
        values = {level: index.get_level_values(level) for level in levels}
        it = DataFrame(values).drop_duplicates().iterrows()

        for k, (i, s) in enumerate(it):
            if not isinstance(i, int):
                raise NotImplementedError

            yield i + offset + k * padding, s.to_dict()


def xs(
    df: DataFrame,
    index: dict[Hashable, Any] | None,
    columns: dict[Hashable, Any] | None,
) -> DataFrame:
    if index:
        for key, value in index.items():
            df = df.xs(value, level=key, axis=0)  # type: ignore

    if columns:
        for key, value in columns.items():
            df = df.xs(value, level=key, axis=1)  # type: ignore

    return df
